from typing import List, Dict, Any, Callable
import random
import numpy as np
from collections import Counter
from csle_common.logging.log import Logger
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_agents.agents.pomcp.node import Node
import csle_agents.constants.constants as constants


class POMCPUtil:
    """
    Class with utility functions related to POMCP
    """

    @staticmethod
    def sample_from_distribution(probability_vector: List[float]) -> int:
        """
        Utility function to sample from a probability vector

        :param probability_vector: the probability vector to sample from
        :return: the sampled element
        """
        probability_vector_np = np.array(probability_vector)
        sample = np.random.choice(list(range(len(probability_vector_np))),
                                  p=probability_vector_np / probability_vector_np.sum())
        return int(sample)

    @staticmethod
    def rand_choice(candidates: List[Any]) -> Any:
        """
        Selects an element from a given list uniformly at random

        :param candidates: the list to sample from
        :return: the sample
        """
        return random.choice(candidates)

    @staticmethod
    def convert_samples_to_distribution(samples) -> Dict[int, float]:
        """
        Converts a list of samples to a probability distribution

        :param samples: the list of samples
        :return: a dict with the sample values and their probabilities
        """
        cnt = Counter(samples)
        _sum = sum(cnt.values())
        return {k: v / _sum for k, v in cnt.items()}

    @staticmethod
    def ucb(history_visit_count, action_visit_count):
        """
        Implements the upper-confidence-bound acquisiton function

        :param history_visit_count: counter of the number of times the history has been visited
        :param action_visit_count: counter of the number of times the action has been taken in the history
        :return: the ucb acquisition value
        """
        # If we have never seen this history before, its exploration utility is initialized to zero
        if history_visit_count == 0:
            return 0.0
        # If we have never taken this action before, its utility is infinity to encourage exploration
        if action_visit_count == 0:
            return np.inf
        # If we have taken this action before, return the UCB exploration bonus
        return np.sqrt(np.log(history_visit_count) / action_visit_count)

    @staticmethod
    def ucb_acquisition_function(action: "Node", c: float) -> float:
        """
        The UCB acquisition function

        :param action: the action node
        :param c: the exploration parameter
        :param rollout_policy: the rollout policy
        :param prior_weight: the weight to put on the prior
        :return: the acquisition value of the action
        """
        if action.parent.visit_count == 0:
            return 0
        if action.visit_count == 0:
            return np.inf
        return float(action.value + c * POMCPUtil.ucb(action.parent.visit_count, action.visit_count))

    @staticmethod
    def alpha_go_acquisition_function(action: "Node", c: float, c2: float, prior: float, prior_weight: float) -> float:
        """
        The UCB acquisition function

        :param action: the action node
        :param c: the exploration parameter
        :param c2: the c2 parameter
        :param prior: the prior weight
        :param prior_weight: the weight to put on the prior
        :return: the acquisition value of the action
        """
        # prior = rollout_policy.probability(o=o, a=action.action)
        # visit_term = math.sqrt(action.parent.visit_count) / (action.visit_count + 1)
        # base_term = math.log((action.parent.visit_count + c2 + 1) / c2 + c)
        # prior_term = prior_weight * prior * visit_term * base_term
        exploration_term = POMCPUtil.ucb(action.parent.visit_count, action.visit_count)
        return float(action.value + (c + prior_weight * prior) * exploration_term)
        # return float(action.value + prior_term + exploration_term)

    @staticmethod
    def trajectory_simulation_particles(o: int, env: BaseEnv, action_sequence: List[int], num_particles: int,
                                        verbose: bool = False) -> List[int]:
        """
        Performs trajectory simulations to find possible states matching to the given observation

        :param o: the observation to match against
        :param env: the black-box simulator to sue for generating trajectories
        :param action_sequence: the action sequence for the trajectory
        :param num_particles: the number of particles to collect
        :param verbose: boolean flag indicating whether logging should be verbose or not
        :return: the list of particles matching the given observation
        """
        particles: List[int] = []
        if verbose:
            Logger.__call__().get_logger().info(f"Filling {num_particles} particles"
                                                f" through trajectory simulations, "
                                                f"action sequence: {action_sequence}, observation: {o}")
        while len(particles) < num_particles:
            done = False
            _, info = env.reset()
            t = 0
            while not done and t < len(action_sequence):
                _, r, done, _, info = env.step(action=action_sequence[t])
                sampled_o = info[constants.COMMON.OBSERVATION]
                s = info[constants.COMMON.STATE]
                if t == len(action_sequence) - 1 and sampled_o == o:
                    particles.append(s)
                t += 1
        return particles

    @staticmethod
    def get_default_value(particles: List[int], action: int, default_value: float, env: BaseEnv,
                          value_function: Callable[[Any], float]) -> float:
        """
        Gets the default value of a node

        :param particles: the particles of the parent node
        :param action: the action of the node
        :param default_value: the default value
        :param env: the black-box simulator
        :param value_function: the value function
        :return: the value
        """
        node_value = default_value
        if value_function is not None:
            sample_values = []
            for i in range(20):
                state = int(POMCPUtil.rand_choice(particles))
                env.set_state(state=state)
                o, r, _, _, info = env.step(action)
                sample_values.append(value_function(o))
            node_value = float(np.mean(sample_values))
        return node_value



from typing import Union, List
import numpy as np
import sys
from typing import Tuple


class GLSUtils:
    """
    Class with util functions for MCS
    """

    def lsrange(self, xl, xu, x, p, prt, bend):
        """
        Defining line search range

        :param xl: lower bound
        :param xu: upper bound
        :param x: starting point
        :param p: Search direction
        :param prt: print command - unused in this implementation sofar
        :param bend:
        """
        if np.max(np.abs(p)) == 0:
            sys.exit('GLS Error: zero search direction in line search')

        # Find sensible step size scale
        if type(p) is not np.ndarray:
            if type(p) is not list:
                p = [p]
            p = np.asarray(p)

        if type(x) is not np.ndarray:
            if type(x) is not list:
                x = [x]
                xl = [xl]
                xu = [xu]
            x = np.asarray(x)
            xl = np.asarray(xl)
            xu = np.asarray(xu)

        # this is test for python
        if x.shape != p.shape:
            sys.exit('GLS Error: dim of x and p does not match: program is going to fail')

        pp = np.abs(p[p != 0])
        u = np.divide(np.abs(x[p != 0]), pp)
        scale = min(u)

        if scale == 0:
            u[u == 0] = np.divide(1, pp[u == 0])
            scale = min(u)

        if not bend:
            # find range of useful alp in truncated line search
            amin = -np.Inf
            amax = np.Inf
            for i in range(len(x)):
                if p[i] > 0:
                    amin = max(amin, (xl[i] - x[i]) / p[i])
                    amax = min(amax, (xu[i] - x[i]) / p[i])
                elif p[i] < 0:
                    amin = max(amin, (xu[i] - x[i]) / p[i])
                    amax = min(amax, (xl[i] - x[i]) / p[i])

            if amin > amax:
                sys.exit('GLS Error: no admissible step in line search')

        else:
            amin = np.Inf
            amax = -np.Inf
            for i in range(len(x)):
                if p[i] > 0:
                    amin = min(amin, (xl[i] - x[i]) / p[i])
                    amax = max(amax, (xu[i] - x[i]) / p[i])
                elif p[i] < 0:
                    amin = min(amin, (xu[i] - x[i]) / p[i])
                    amax = max(amax, (xl[i] - x[i]) / p[i])

        return xl, xu, x, p, amin, amax, scale

    def lssort(self, alist: List[Union[float, int]], flist: List[Union[float, int]]):
        """
        Performing the lssort

        :param alist: list of known steps
        :param flist: function values of known steps
        :return: metrics and parameters obtained from doing the lssort
        """
        perm = np.argsort(alist).tolist()
        alist.sort()
        flist = [flist[i] for i in perm if i < len(flist)]
        if len(flist) >= len(alist):
            s = len(alist)
        else:
            s = len(flist)

        up = [i < j for i, j in zip(flist[0: s - 1], flist[1: s])]
        down = [i <= j for i, j in zip(flist[1: s], flist[0: s - 1])]
        if len(down) == 1:
            down[0] = flist[s - 1] < flist[s - 2]
        else:
            down[s - 2] = flist[s - 1] < flist[s - 2]

        monotone = (sum(up) == 0 or sum(down) == 0)
        minima = [i and j for i, j in zip(up + [True], [True] + down)]
        nmin = sum(minima)

        fbest = min(flist)
        i = np.argmin(flist)

        abest = alist[i]
        fmed = np.median(flist)

        if nmin > 1:
            al = [alist[i] for i in range(len(minima)) if minima[i]]
            if abest in al:
                al.remove(abest)
            unitlen = min(np.abs(np.subtract(al, abest)))
            # TODO: investigate edge case error
        else:
            unitlen = max(abest - alist[0], alist[s - 1] - abest)

        return alist, flist, abest, fbest, fmed, up, down, monotone, minima, nmin, unitlen, s

    def lsconvex(self, alist: List[Union[float, int]], flist: List[Union[float, int]],
                 nmin: int, s: int) -> int:
        """
        Performing the lsconvex
        :param alist: list of known steps
        :param flist: function values of known steps
        :param nmin:
        :param s:
        :return: convex
        """
        if nmin > 1:
            convex = 0
        else:
            convex = 1
            for i in range(1, s - 1):
                f12 = (flist[i] - flist[i - 1]) / (alist[i] - alist[i - 1])
                f13 = (flist[i] - flist[i + 1]) / (alist[i] - alist[i + 1])
                f123 = (f13 - f12) / (alist[i + 1] - alist[i - 1])
                if f123 < 0:
                    convex = 0
                    break
            # if convex:
            #     nothing_to_do = 'done!'
        return convex

    def lssat(self, small: Union[float, int], alist: List[Union[float, int]],
              flist: List[Union[float, int]], alp: float, amin: float, amax: float, s: int,
              saturated: int) -> Tuple[float, int]:
        """
        Performing the lssat
        :param small:
        :param alist:
        :param flist:
        :param alp:
        :param amin:
        :param amin:
        :param amax:
        :param s:
        :param saturated:
        :return: alp, saturated
        """
        cont = saturated
        if cont:
            i = np.argmin(flist)
            if i == 0 or i == s - 1:
                cont = 0

        if cont:
            aa = [alist[j] for j in range(i - 1, i + 1 + 1)]
            ff = [flist[j] for j in range(i - 1, i + 1 + 1)]

            f12 = (ff[1] - ff[0]) / (aa[1] - aa[0])
            f23 = (ff[2] - ff[1]) / (aa[2] - aa[1])
            f123 = (f23 - f12) / (aa[2] - aa[0])

            if f123 > 0:
                alp = 0.5 * (aa[1] + aa[2] - f23 / f123)
                alp = max(amin, min(alp, amax))
                alptol = small * (aa[2] - aa[0])
                saturated = (abs(alist[i] - alp) <= alptol)
            else:
                saturated = 0
            # if not saturated:
            #     no_print = 0
        return alp, saturated



"""
MIT License

Copyright (c) 2019 CleanRL developers https://github.com/vwxyzjn/cleanrl
"""

from typing import Union, List, Optional, Callable, Tuple, Any
import random
import time
import gymnasium as gym
from gymnasium.wrappers.common import RecordEpisodeStatistics
from gymnasium.spaces.discrete import Discrete
import os
import numpy as np
import torch
import torch.nn.utils.clip_grad as clip_grad
import torch.optim as optim
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.util.general_util import GeneralUtil
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_common.dao.training.ppo_policy import PPOPolicy
from csle_agents.agents.base.base_agent import BaseAgent
from csle_common.models.ppo_network import PPONetwork
import csle_agents.constants.constants as agents_constants


class PPOCleanAgent(BaseAgent):
    """
    A PPO agent using the implementation from CleanRL
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 emulation_env_config: Union[None, EmulationEnvConfig], experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True) -> None:
        """
        Initializes the agent, and sets the hyperparameters as attributes of the class representing the agent.

        :param simulation_env_config: the simulation environment configuration
        :param emulation_env_config: the emulation environment configuration
        :param experiment_config: the experiment configuration
        :param training_job: the training job
        :param save_to_metastore: boolean flag indicating whether the results should be saved to the metastore or not
        """
        super(PPOCleanAgent, self).__init__(simulation_env_config=simulation_env_config,
                                            emulation_env_config=emulation_env_config,
                                            experiment_config=experiment_config)
        assert experiment_config.agent_type == AgentType.PPO_CLEAN
        self.training_job = training_job
        self.save_to_metastore = save_to_metastore
        self.num_steps = self.experiment_config.hparams[agents_constants.PPO_CLEAN.NUM_STEPS].value
        self.num_envs = self.experiment_config.hparams[agents_constants.COMMON.NUM_PARALLEL_ENVS].value
        self.batch_size = self.experiment_config.hparams[agents_constants.COMMON.BATCH_SIZE].value
        self.total_timesteps = self.experiment_config.hparams[
            agents_constants.COMMON.NUM_TRAINING_TIMESTEPS].value
        self.num_minibatches = self.experiment_config.hparams[agents_constants.PPO_CLEAN.NUM_MINIBATCHES].value
        self.minibatch_size = int(self.batch_size // self.num_minibatches)
        self.num_iterations = self.total_timesteps // self.batch_size
        self.update_epochs = self.experiment_config.hparams[agents_constants.PPO_CLEAN.UPDATE_EPOCHS].value
        self.clip_coef = self.experiment_config.hparams[agents_constants.PPO_CLEAN.CLIP_RANGE].value
        self.clip_vloss = self.experiment_config.hparams[agents_constants.PPO_CLEAN.CLIP_VLOSS].value
        self.norm_adv = self.experiment_config.hparams[agents_constants.PPO_CLEAN.NORM_ADV].value
        self.vf_coef = self.experiment_config.hparams[agents_constants.PPO_CLEAN.CLIP_RANGE_VF].value
        self.ent_coef = self.experiment_config.hparams[agents_constants.PPO_CLEAN.ENT_COEF].value
        self.max_grad_norm = self.experiment_config.hparams[agents_constants.PPO_CLEAN.MAX_GRAD_NORM].value
        self.target_kl = None
        if self.experiment_config.hparams[agents_constants.PPO_CLEAN.TARGET_KL].value != -1:
            self.target_kl = self.experiment_config.hparams[agents_constants.PPO_CLEAN.TARGET_KL].value
        self.anneal_lr = self.experiment_config.hparams[agents_constants.PPO_CLEAN.ANNEAL_LR].value
        self.gamma = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        self.gae_lambda = self.experiment_config.hparams[agents_constants.PPO_CLEAN.GAE_LAMBDA].value
        self.learning_rate = self.experiment_config.hparams[agents_constants.COMMON.LEARNING_RATE].value
        config = self.simulation_env_config.simulation_env_input_config
        self.orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=config)

    def train(self) -> ExperimentExecution:
        """
        Runs the training process

        :return: the results
        """
        pid = os.getpid()

        # Setup experiment metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RANDOM_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNTIME)
        descr = f"Training of policies with Clean-PPO using " \
                f"simulation:{self.simulation_env_config.name}"

        # Setup training job
        if self.training_job is None:
            emulation_name = ""
            if self.emulation_env_config is not None:
                emulation_name = self.emulation_env_config.name
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name=emulation_name, simulation_traces=[],
                num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            training_job_id = -1
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
            self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Setup experiment execution
        ts = time.time()
        emulation_name = ""
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        simulation_name = self.simulation_env_config.name
        self.exp_execution = ExperimentExecution(
            result=exp_result, config=self.experiment_config, timestamp=ts,
            emulation_name=emulation_name, simulation_name=simulation_name, descr=descr,
            log_file_path=self.training_job.log_file_path)
        exp_execution_id = -1
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
        self.exp_execution.id = exp_execution_id

        # Training runs, one per seed
        for seed in self.experiment_config.random_seeds:

            # Train
            exp_result, env, model = self.run_ppo(exp_result=exp_result, seed=seed)

            # Save policy
            ts = time.time()
            save_path = f"{self.experiment_config.output_dir}/ppo_policy_seed_{seed}_{ts}.zip"
            model.save(save_path)
            policy = PPOPolicy(
                model=model, simulation_name=self.simulation_env_config.name, save_path=save_path,
                states=self.simulation_env_config.state_space_config.states,
                actions=self.simulation_env_config.joint_action_space_config.action_spaces[
                    self.experiment_config.player_idx].actions, player_type=self.experiment_config.player_type,
                experiment_config=self.experiment_config,
                avg_R=exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN][-1])
            exp_result.policies[seed] = policy

            # Save policy metadata
            if self.save_to_metastore:
                MetastoreFacade.save_ppo_policy(ppo_policy=policy)
                os.chmod(save_path, 0o777)

            # Save trace
            traces = env.get_traces()
            if len(traces) > 0 and self.save_to_metastore:
                MetastoreFacade.save_simulation_trace(traces[-1])
            env.reset_traces()

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                seed_values = []
                for seed_idx in range(len(self.experiment_config.random_seeds)):
                    seed_values.append(value_vectors[seed_idx][i])
                avg_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0])
                std_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1])
            exp_result.avg_metrics[metric] = avg_metrics
            exp_result.std_metrics[metric] = std_metrics
        return self.exp_execution

    def run_ppo(self, exp_result: ExperimentResult, seed: int) -> Tuple[ExperimentResult, BaseEnv, PPONetwork]:
        """
        Runs PPO with a given seed

        :param exp_result: the object to save the experiment results
        :param seed: the random seed
        :return: the updated experiment results, the environment, and the trained model
        """
        Logger.__call__().get_logger().info(f"[CleanPPO] Start training; seed: {seed}")

        envs = gym.vector.SyncVectorEnv([self.make_env() for _ in range(self.num_envs)])

        # Setup training metrics
        exp_result.all_metrics[seed] = {}
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RANDOM_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME] = []
        ExperimentUtil.set_seed(seed)
        cuda = False

        # Create neural network
        device = torch.device(agents_constants.PPO_CLEAN.CUDA if torch.cuda.is_available() and cuda else
                              self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value)
        num_hidden_layers = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS].value
        hidden_layer_dim = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER].value
        input_dim = np.array(envs.single_observation_space.shape).prod()
        env: BaseEnv = self.orig_env
        action_space: Discrete = env.action_space
        action_dim = int(action_space.n)
        model = PPONetwork(input_dim=input_dim, output_dim_critic=1, output_dim_action=action_dim,
                           num_hidden_layers=num_hidden_layers, hidden_layer_dim=hidden_layer_dim).to(device)

        # seeding
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = True

        # Setup gym environment
        obs_shape: Tuple[int, ...] = (1,)
        action_shape: Tuple[int, ...] = (1,)
        if envs.single_observation_space.shape is not None:
            obs_shape = envs.single_observation_space.shape
        if envs.single_action_space.shape is not None:
            action_shape = envs.single_action_space.shape
        obs = torch.zeros((self.num_steps, self.num_envs) + obs_shape).to(device)
        actions = torch.zeros((self.num_steps, self.num_envs) + action_shape).to(device)
        logprobs = torch.zeros((self.num_steps, self.num_envs)).to(device)
        rewards = torch.zeros((self.num_steps, self.num_envs)).to(device)
        dones = torch.zeros((self.num_steps, self.num_envs)).to(device)
        values = torch.zeros((self.num_steps, self.num_envs)).to(device)
        horizons: List[int] = []

        # Initialize the environment and optimizers
        global_step = 0
        start_time = time.time()
        next_obs = torch.Tensor(envs.reset(seed=seed)[0]).to(device)
        next_done = torch.zeros(self.num_envs).to(device)
        optimizer = optim.Adam(model.parameters(), lr=self.learning_rate, eps=1e-5)

        # Training
        for iteration in range(1, self.num_iterations + 1):
            # Annealing the rate if instructed to do so.
            if self.anneal_lr:
                frac = 1.0 - (iteration - 1.0) / self.num_iterations
                lrnow = frac * self.learning_rate
                optimizer.param_groups[0]["lr"] = lrnow
            next_obs, next_done, global_step, horizons = \
                self.update_trajectory_buffers(global_step=global_step, envs=envs, obs=obs, dones=dones,
                                               actions=actions, rewards=rewards, device=device,
                                               logprobs=logprobs, values=values, model=model, next_obs=next_obs,
                                               next_done=next_done, horizons=horizons)

            returns, advantages = self.generalized_advantage_estimation(model=model, next_obs=next_obs, rewards=rewards,
                                                                        device=device, next_done=next_done,
                                                                        dones=dones, values=values)

            # flatten the batch
            b_obs = obs.reshape((-1,) + obs_shape)
            b_logprobs = logprobs.reshape(-1)
            b_actions = actions.reshape((-1,) + action_shape)
            b_advantages = advantages.reshape(-1)
            b_returns = returns.reshape(-1)
            b_values = values.reshape(-1)

            # Optimizing the policy and value network
            b_inds = np.arange(self.batch_size)
            clipfracs = []
            for epoch in range(self.update_epochs):
                np.random.shuffle(b_inds)
                for start in range(0, self.batch_size, self.minibatch_size):
                    end = start + self.minibatch_size
                    mb_inds = list(b_inds[start:end])

                    _, newlogprob, entropy, newvalue = \
                        model.get_action_and_value(b_obs[mb_inds], b_actions.long()[mb_inds])
                    logratio = newlogprob - b_logprobs[mb_inds]
                    ratio = logratio.exp()

                    with torch.no_grad():
                        # calculate approx_kl http://joschu.net/blog/kl-approx.html
                        approx_kl = ((ratio - 1) - logratio).mean()
                        clipfracs += [((ratio - 1.0).abs() > self.clip_coef).float().mean().item()]

                    mb_advantages = b_advantages[mb_inds]
                    if self.norm_adv:
                        mb_advantages = (mb_advantages - mb_advantages.mean()) / (mb_advantages.std() + 1e-8)

                    # Policy loss
                    pg_loss1 = -mb_advantages * ratio
                    pg_loss2 = -mb_advantages * torch.clamp(ratio, 1 - self.clip_coef, 1 + self.clip_coef)
                    pg_loss = torch.max(pg_loss1, pg_loss2).mean()

                    # Value loss
                    newvalue = newvalue.view(-1)
                    if self.clip_vloss:
                        v_loss_unclipped = (newvalue - b_returns[mb_inds]) ** 2
                        v_clipped = b_values[mb_inds] + torch.clamp(
                            newvalue - b_values[mb_inds],
                            -self.clip_coef,
                            self.clip_coef,
                        )
                        v_loss_clipped = (v_clipped - b_returns[mb_inds]) ** 2
                        v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                        v_loss = 0.5 * v_loss_max.mean()
                    else:
                        v_loss = 0.5 * ((newvalue - b_returns[mb_inds]) ** 2).mean()

                    # Entropy loss
                    entropy_loss = entropy.mean()

                    # Total loss
                    loss = pg_loss - self.ent_coef * entropy_loss + v_loss * self.vf_coef

                    # Backpropagation and update weights
                    optimizer.zero_grad()
                    loss.backward()
                    clip_grad.clip_grad_norm_(model.parameters(), self.max_grad_norm)
                    optimizer.step()

                if self.target_kl is not None and approx_kl > self.target_kl:
                    break

            # Logging
            time_elapsed_minutes = round((time.time() - start_time) / 60, 3)
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
            avg_R = round(float(np.mean(returns.flatten().cpu().numpy())), 3)
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN].append(round(avg_R, 3))
            avg_T = float(np.mean(horizons))
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON].append(
                round(avg_T, 3))
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
            running_avg_J = ExperimentUtil.running_average(
                exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN],
                self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN].append(
                round(running_avg_J, 3))
            running_avg_T = ExperimentUtil.running_average(
                exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON],
                self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON].append(
                round(running_avg_T, 3))
            Logger.__call__().get_logger().info(
                f"[CleanPPO] Iteration: {iteration}/{self.num_iterations}, "
                f"avg R: {avg_R}, "
                f"R_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}:"
                f"{running_avg_J}, Avg T:{round(avg_T, 3)}, "
                f"Running_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}_T: "
                f"{round(running_avg_T, 3)}, "
                f"runtime: {time_elapsed_minutes} min")

        envs.close()
        base_env: BaseEnv = envs.envs[0].env.env.env  # type: ignore
        return exp_result, base_env, model

    def make_env(self) -> Callable[[], RecordEpisodeStatistics[Any, Any]]:
        """
        Helper function for creating the environment to use for training

        :return: a function that creates the environment
        """

        def thunk() -> RecordEpisodeStatistics[Any, Any]:
            """
            Function for creating a new environment

            :return: the created environment
            """
            config = self.simulation_env_config.simulation_env_input_config
            orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=config)
            env = RecordEpisodeStatistics(orig_env)
            return env

        return thunk

    def update_trajectory_buffers(self, global_step: int, envs: gym.vector.SyncVectorEnv, obs: torch.Tensor,
                                  dones: torch.Tensor, actions: torch.Tensor, rewards: torch.Tensor,
                                  device: torch.device, logprobs: torch.Tensor, values: torch.Tensor,
                                  model: PPONetwork, next_obs: torch.Tensor, next_done: torch.Tensor,
                                  horizons: List[int]) \
            -> Tuple[torch.Tensor, torch.Tensor, int, List[int]]:
        """
        Updates the buffers of trajectories collected from the environment

        :param global step: the global step in the iteration
        :param envs: list of environments
        :param obs: torch tensor of observations
        :param dones: tensor of done events
        :param actions: tensor of available actions
        :param rewards: tensor of available rewards
        :param device: the device acted upon
        :param logprobs: logarithmic probabilities
        :param horizons: list of time horizons
        :param values: tensor of values
        :param model: the neural network model
        :param next_obs: the next observation
        :param next_done: logical or operation between terminations and truncations
        :return: next_obs, next_done, global_step
        """
        step_dones = [0] * self.num_envs
        for step in range(0, self.num_steps):
            global_step += self.num_envs
            obs[step] = next_obs
            dones[step] = next_done
            for i, d in enumerate(next_done.cpu().numpy()):
                if float(d) == 1.0:
                    horizons.append(step - step_dones[i])
                    step_dones[i] = step
            with torch.no_grad():
                action, logprob, _, value = model.get_action_and_value(next_obs)
                values[step] = value.flatten()
            actions[step] = action
            logprobs[step] = logprob

            # Step the environment
            next_obs, reward, terminations, truncations, infos = envs.step(action.cpu().numpy()) # type: ignore
            next_done_np = np.logical_or(terminations, truncations)
            rewards[step] = torch.tensor(reward).to(device).view(-1)
            next_obs, next_done = torch.Tensor(next_obs).to(device), torch.Tensor(next_done_np).to(device)
        return next_obs, next_done, global_step, horizons

    def generalized_advantage_estimation(self, model: PPONetwork, next_obs: torch.Tensor, rewards: torch.Tensor,
                                         device: torch.device, next_done: torch.Tensor, dones: torch.Tensor,
                                         values: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Computes the generalized advantage estimation (i.e., exponentially weighted average of n-step returns)
        See (HIGH-DIMENSIONAL CONTINUOUS CONTROL USING GENERALIZED ADVANTAGE ESTIMATION, 2016, ICLR)

        :param device: the device acted upon
        :param values: tensor of values
        :param model: the neural network model
        :param rewards: tensor of available rewards
        :param next_obs: the next observation
        :param dones: tensor of done events
        :param next_done: logical or operation between terminations and truncations
        :return: returns, advantages
        """
        with torch.no_grad():
            next_value = model.get_value(next_obs).reshape(1, -1)
            advantages = torch.zeros_like(rewards).to(device)
            lastgaelam = 0
            for t in reversed(range(self.num_steps)):
                if t == self.num_steps - 1:
                    nextnonterminal = 1.0 - next_done
                    nextvalues = next_value
                else:
                    nextnonterminal = 1.0 - dones[t + 1]
                    nextvalues = values[t + 1]
                delta = rewards[t] + self.gamma * nextvalues * nextnonterminal - values[t]
                advantages[t] = lastgaelam = delta + self.gamma * self.gae_lambda * nextnonterminal * lastgaelam
            returns = advantages + values
            return returns, advantages

    def hparam_names(self) -> List[str]:
        """
        :return: a list with the hyperparameter names
        """
        return [constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER,
                constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS,
                agents_constants.PPO_CLEAN.NUM_MINIBATCHES,
                agents_constants.COMMON.NUM_PARALLEL_ENVS, agents_constants.COMMON.BATCH_SIZE,
                agents_constants.COMMON.NUM_TRAINING_TIMESTEPS, agents_constants.PPO_CLEAN.GAE_LAMBDA,
                agents_constants.PPO_CLEAN.CLIP_RANGE, agents_constants.COMMON.LEARNING_RATE,
                agents_constants.PPO_CLEAN.NORM_ADV, agents_constants.COMMON.GAMMA,
                agents_constants.PPO_CLEAN.CLIP_RANGE_VF, agents_constants.PPO_CLEAN.ENT_COEF,
                agents_constants.PPO_CLEAN.VF_COEF, agents_constants.PPO_CLEAN.UPDATE_EPOCHS,
                agents_constants.PPO_CLEAN.TARGET_KL, agents_constants.PPO_CLEAN.MAX_GRAD_NORM,
                agents_constants.COMMON.EVAL_EVERY, constants.NEURAL_NETWORKS.DEVICE,
                agents_constants.PPO_CLEAN.CLIP_VLOSS, agents_constants.PPO_CLEAN.ANNEAL_LR,
                agents_constants.COMMON.SAVE_EVERY]



"""
MIT License

Copyright (c) 2019 CleanRL developers https://github.com/vwxyzjn/cleanrl
"""
from typing import List, Optional, Tuple, Any
import time
import torch
import torch.optim as optim
import gymnasium as gym
import os
import random
import numpy as np
from stable_baselines3.common.buffers import ReplayBuffer
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from gymnasium.wrappers.common import RecordEpisodeStatistics
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.models.q_network import QNetwork
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.dao.training.dqn_policy import DQNPolicy
from csle_common.util.general_util import GeneralUtil
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_agents.agents.base.base_agent import BaseAgent
import csle_agents.constants.constants as agents_constants


class C51CleanAgent(BaseAgent):
    """
    A C51 agent using the implementation from CleanRL documentation
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 emulation_env_config: Optional[EmulationEnvConfig], experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True) -> None:
        """
        Initializes the agent

        :param simulation_env_config: the simulation environment configuration
        :param emulation_env_config: the emulation environment configuration
        :param experiment_config: the experiment configuration
        :param training_job: the training job configuration
        :param save_to_metastore: boolean flag indicating whether job information should be saved to the metastore
        """
        super(C51CleanAgent, self).__init__(simulation_env_config=simulation_env_config,
                                            emulation_env_config=emulation_env_config,
                                            experiment_config=experiment_config)
        assert experiment_config.agent_type == AgentType.C51_CLEAN
        self.training_job = training_job
        self.num_hl = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS].value
        self.num_hl_neur = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER].value
        self.config = self.simulation_env_config.simulation_env_input_config
        self.save_to_metastore = save_to_metastore
        self.exp_name: str = self.simulation_env_config.name
        self.env_id = self.simulation_env_config.gym_env_name
        self.torch_deterministic = True
        self.cuda: bool = True
        self.learning_rate = self.experiment_config.hparams[agents_constants.COMMON.LEARNING_RATE].value
        self.num_envs = self.experiment_config.hparams[agents_constants.COMMON.NUM_PARALLEL_ENVS].value
        self.total_timesteps = self.experiment_config.hparams[
            agents_constants.COMMON.NUM_TRAINING_TIMESTEPS].value

        self.batch_size = self.experiment_config.hparams[agents_constants.COMMON.BATCH_SIZE].value
        self.num_iterations = self.total_timesteps // self.batch_size
        self.gamma = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        self.tau = self.experiment_config.hparams[agents_constants.C51_CLEAN.TAU].value
        self.exploration_fraction = self.experiment_config.hparams[agents_constants.C51_CLEAN.EXP_FRAC].value
        self.learning_starts = self.experiment_config.hparams[agents_constants.C51_CLEAN.LEARNING_STARTS].value
        self.train_frequency = self.experiment_config.hparams[agents_constants.C51_CLEAN.TRAIN_FREQ].value
        self.target_network_frequency = self.experiment_config.hparams[agents_constants.C51_CLEAN.T_N_FREQ].value
        self.buffer_size = self.experiment_config.hparams[agents_constants.C51_CLEAN.BUFFER_SIZE].value
        self.save_model = self.experiment_config.hparams[agents_constants.C51_CLEAN.SAVE_MODEL].value
        self.device = self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value
        self.start_exploration_rate = (
            self.experiment_config.hparams[agents_constants.C51_CLEAN.START_EXPLORATION_RATE].value)
        self.end_eploration_rate = (
            self.experiment_config.hparams[agents_constants.C51_CLEAN.END_EXPLORATION_RATE].value)
        self.n_atoms = self.experiment_config.hparams[agents_constants.C51_CLEAN.N_ATOMS].value
        self.v_min = self.experiment_config.hparams[agents_constants.C51_CLEAN.V_MIN].value
        self.v_max = self.experiment_config.hparams[agents_constants.C51_CLEAN.V_MAX].value
        self.orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=self.config)

    def linear_schedule(self, duration: int, t: int) -> float:
        """
        Linear exploration rate schedule

        :param duration: training duration
        :param t: the time step of the training
        :return: the updated exploration rate
        """
        slope = (self.end_eploration_rate - self.start_exploration_rate) / duration
        return float(max(slope * t + self.start_exploration_rate, self.end_eploration_rate))

    def make_env(self):
        """
        Helper function for creating an environment in training of the agent

        :return: environment creating function
        """

        def thunk():
            """
            The environment creating function
            """
            orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=self.config)
            env = RecordEpisodeStatistics(orig_env)
            return env

        return thunk

    def train(self) -> ExperimentExecution:
        """
        Implements the training logic of the C51 algorithm

        :return: the experiment result
        """
        pid = os.getpid()

        # Setup experiment metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RANDOM_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNTIME)
        descr = f"Training of policies with C51 using " \
                f"simulation:{self.simulation_env_config.name}"

        # Setup training job
        if self.training_job is None:
            emulation_name = ""
            if self.emulation_env_config is not None:
                emulation_name = self.emulation_env_config.name
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name=emulation_name, simulation_traces=[],
                num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            training_job_id = -1
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
            self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Setup experiment execution
        ts = time.time()
        simulation_name = self.simulation_env_config.name
        emulation_name = ""
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        self.exp_execution = ExperimentExecution(
            result=exp_result, config=self.experiment_config, timestamp=ts,
            emulation_name=emulation_name, simulation_name=simulation_name,
            descr=descr, log_file_path=self.training_job.log_file_path)
        exp_execution_id = -1
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
        self.exp_execution.id = exp_execution_id

        # Training runs, one per seed
        for seed in self.experiment_config.random_seeds:
            assert self.num_envs == 1, "vectorized envs are not supported at the moment"

            # Train
            exp_result, env, model = self.run_c51(exp_result=exp_result, seed=seed)

            # Save policy
            ts = time.time()
            save_path = f"{self.experiment_config.output_dir}/C51_policy_seed_{seed}_{ts}.zip"
            model.save(save_path)
            action_space = \
                self.simulation_env_config.joint_action_space_config.action_spaces[
                    self.experiment_config.player_idx].actions

            policy = DQNPolicy(model=model, simulation_name=self.simulation_env_config.name,
                               save_path=save_path, player_type=self.experiment_config.player_type,
                               states=self.simulation_env_config.state_space_config.states,
                               actions=action_space,
                               experiment_config=self.experiment_config,
                               avg_R=exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN][-1]
                               )

            if self.save_to_metastore:
                MetastoreFacade.save_dqn_policy(dqn_policy=policy)
                os.chmod(save_path, 0o777)

            # Save trace
            traces = env.get_traces()
            if len(traces) > 0 and self.save_to_metastore:
                MetastoreFacade.save_simulation_trace(traces[-1])
            env.reset_traces()

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                seed_values = []
                for seed_idx in range(len(self.experiment_config.random_seeds)):
                    seed_values.append(value_vectors[seed_idx][i])
                avg_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0])
                std_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1])
            exp_result.avg_metrics[metric] = avg_metrics
            exp_result.std_metrics[metric] = std_metrics
        return self.exp_execution

    def run_c51(self, exp_result: ExperimentResult, seed: int) -> Tuple[ExperimentResult, BaseEnv, QNetwork]:
        """
        Runs C51 with given seed

        :param exp_result: the object to save the experiment results
        :param seed: the random seed
        :retur: the updated experiment results, the environment and the trained model
        """
        Logger.__call__().get_logger().info(f"[CleanC51] Start training; seed: {seed}")
        envs = gym.vector.SyncVectorEnv([self.make_env() for _ in range(self.num_envs)])
        assert isinstance(envs.single_action_space, gym.spaces.Discrete), "only discrete action space is supported"

        # Setup training metrics
        exp_result.all_metrics[seed] = {}
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RANDOM_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME] = []
        ExperimentUtil.set_seed(seed)
        cuda = False

        # Create neural network
        device = torch.device(agents_constants.C51_CLEAN.CUDA if torch.cuda.is_available() and cuda else
                              self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value)
        input_dim = np.array(envs.single_observation_space.shape).prod()
        q_network = QNetwork(input_dim, num_hidden_layers=self.num_hl, hidden_layer_dim=self.num_hl_neur,
                             agent_type=self.experiment_config.agent_type, n_atoms=self.n_atoms,
                             action_space_dim=envs.single_action_space.n).to(device)
        optimizer = optim.Adam(q_network.parameters(), lr=self.learning_rate, eps=0.01 / self.batch_size)
        target_network = QNetwork(
            input_dim, num_hidden_layers=self.num_hl, hidden_layer_dim=self.num_hl_neur,
            agent_type=self.experiment_config.agent_type, n_atoms=self.n_atoms,
            action_space_dim=envs.single_action_space.n).to(device)

        # Seeding
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = self.torch_deterministic

        target_network.load_state_dict(q_network.state_dict())
        rb = ReplayBuffer(self.buffer_size, envs.single_observation_space, envs.single_action_space, device,
                          handle_timeout_termination=False)
        start_time = time.time()
        obs: List[float] = envs.reset(seed=seed)[0]
        R: List[Any] = []
        T = []
        i = 0
        for global_step in range(self.total_timesteps):
            i += 1

            epsilon = self.linear_schedule(duration=self.exploration_fraction * self.total_timesteps, t=global_step)
            if random.random() < epsilon:
                actions = np.array([envs.single_action_space.sample() for _ in range(envs.num_envs)])
            else:
                actions_tensor, pmf = q_network.get_action(torch.Tensor(obs).to(device))
                actions = actions_tensor.cpu().numpy()

            next_obs, rewards, terminations, truncations, infos = envs.step(actions) # type: ignore

            if "final_info" in infos:
                R_sum = 0
                T_sum = 0
                for info in infos["final_info"]:
                    R_sum += info["R"]
                    T_sum += info["T"]
                R.append(R_sum)
                T.append(T_sum)

            real_next_obs = next_obs.copy()
            for idx, trunc in enumerate(truncations):
                if trunc and "final_observation" in infos:
                    real_next_obs[idx] = infos["final_observation"][idx]
            rb.add(obs, real_next_obs, actions, rewards, terminations, infos)  # type: ignore

            obs = next_obs
            if global_step > self.learning_starts:

                if global_step % self.train_frequency == 0:
                    data = rb.sample(self.batch_size)
                    with torch.no_grad():
                        _, next_pmfs = target_network.get_action(data.next_observations.to(dtype=torch.float32))
                        next_atoms = data.rewards + self.gamma * target_network.atoms * (1 - data.dones)

                        # projection
                        delta_z = target_network.atoms[1] - target_network.atoms[0]
                        tz = next_atoms.clamp(self.v_min, self.v_max)

                        b = (tz - self.v_min) / delta_z
                        l = b.floor().clamp(0, self.n_atoms - 1)
                        u = b.ceil().clamp(0, self.n_atoms - 1)

                        d_m_l = (u + (l == u).float() - b) * next_pmfs
                        d_m_u = (b - l) * next_pmfs
                        target_pmfs = torch.zeros_like(next_pmfs)
                        for i in range(target_pmfs.size(0)):
                            target_pmfs[i].index_add_(0, l[i].long(), d_m_l[i])
                            target_pmfs[i].index_add_(0, u[i].long(), d_m_u[i])
                    _, old_pmfs = q_network.get_action(data.observations.to(dtype=torch.float32))
                    loss = (-(target_pmfs * old_pmfs.clamp(min=1e-5, max=1 - 1e-5).log()).sum(-1)).mean()

                    # optimize the model
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()

                # update target network
                if global_step % self.target_network_frequency == 0:
                    target_network.load_state_dict(q_network.state_dict())
            # Logging
            if global_step > 10 and global_step % self.experiment_config.log_every == 0:
                time_elapsed_minutes = round((time.time() - start_time) / 60, 3)
                exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
                avg_R = round(float(np.mean(R)), 3)
                exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN].append(round(avg_R, 3))
                avg_T = round(float(np.mean(T)), 3)
                exp_result.all_metrics[seed][
                    agents_constants.COMMON.AVERAGE_TIME_HORIZON].append(round(avg_T, 3))
                running_avg_J = ExperimentUtil.running_average(
                    exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN],
                    self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value
                )
                running_avg_T = ExperimentUtil.running_average(
                    exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON],
                    self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
                exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN].append(
                    round(running_avg_J, 3))
                Logger.__call__().get_logger().info(
                    f"[C51Clean] Iteration: {global_step}/{self.total_timesteps}, "
                    f"avg R: {avg_R}, "
                    f"R_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}:"
                    f"{running_avg_J}, Avg T:{round(avg_T, 3)}, "
                    f"Running_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}_T: "
                    f"{round(running_avg_T, 3)}, "
                    f"runtime: {time_elapsed_minutes} min")
        envs.close()

        base_env: BaseEnv = envs.envs[0].env.env.env  # type: ignore
        return exp_result, base_env, q_network

    def hparam_names(self) -> List[str]:
        """
        Gets the hyperparameters

        :return: a list with the hyperparameter names
        """
        return [constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER,
                constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS,
                agents_constants.COMMON.LEARNING_RATE, agents_constants.COMMON.BATCH_SIZE,
                agents_constants.COMMON.GAMMA,
                agents_constants.COMMON.NUM_TRAINING_TIMESTEPS, agents_constants.COMMON.EVAL_EVERY,
                agents_constants.COMMON.EVAL_BATCH_SIZE,
                constants.NEURAL_NETWORKS.DEVICE,
                agents_constants.COMMON.SAVE_EVERY,
                agents_constants.C51_CLEAN.LEARNING_STARTS,
                agents_constants.C51_CLEAN.BUFFER_SIZE, agents_constants.C51_CLEAN.N_ATOMS,
                agents_constants.C51_CLEAN.V_MIN, agents_constants.C51_CLEAN.V_MAX,
                agents_constants.COMMON.L, agents_constants.C51_CLEAN.NORM_ADV,
                agents_constants.C51_CLEAN.START_EXPLORATION_RATE, agents_constants.C51_CLEAN.END_EXPLORATION_RATE]



from typing import List, Union, Callable, Any, Dict, Tuple
import time
import random
import torch
import math
import numpy as np
import copy
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_common.dao.training.policy import Policy
from csle_common.logging.log import Logger
from csle_agents.agents.pomcp.belief_tree import BeliefTree
from csle_agents.agents.pomcp.belief_node import BeliefNode
from csle_agents.agents.pomcp.action_node import ActionNode
from csle_agents.agents.pomcp.pomcp_acquisition_function_type import POMCPAcquisitionFunctionType
from csle_agents.agents.pomcp.pomcp_util import POMCPUtil
import csle_agents.constants.constants as constants


class POMCP:
    """
    Class that implements the POMCP algorithm
    """

    def __init__(self, A: List[int], gamma: float, env: BaseEnv, c: float,
                 initial_particles: List[Any], planning_time: float = 0.5, max_particles: int = 350,
                 reinvigoration: bool = False,
                 reinvigorated_particles_ratio: float = 0.1,
                 rollout_policy: Union[Policy, None] = None,
                 value_function: Union[Callable[[Any], float], None] = None, verbose: bool = False,
                 default_node_value: float = 0, prior_weight: float = 1.0, prior_confidence: int = 0,
                 acquisition_function_type: POMCPAcquisitionFunctionType = POMCPAcquisitionFunctionType.UCB,
                 c2: float = 1, use_rollout_policy: bool = False, prune_action_space: bool = False,
                 prune_size: int = 3) -> None:
        """
        Initializes the solver

        :param S: the state space
        :param A: the action space
        :param gamma: the discount factor
        :param env: the environment for sampling
        :param c: the weighting factor for UCB
        :param initial_particles: the initial list of particles
        :param planning_time: the planning time
        :param max_particles: the maximum number of particles (samples) for the belief state
        :param reinvigorated_particles_ratio: probability of new particles added when updating the belief state
        :param reinvigoration: boolean flag indicating whether reinvigoration should be done
        :param rollout_policy: the rollout policy
        :param verbose: boolean flag indicating whether logging should be verbose
        :param default_node_value: the default value of nodes in the tree
        :param num_evals_per_process: number of evaluations per process
        :param prior_weight: the weight to put on the prior
        :param prior_confidence: the prior confidence (initial count)
        :param acquisition_function_type: the type of acquisition function
        :param c2: the weighting factor for alpha go exploration
        :param use_rollout_policy: boolean flag indicating whether rollout policy should be used
        :param prune_action_space: boolean flag indicating whether the action space should be pruned
        :param prune_size: the size of the pruned action space
        """
        self.A = A
        self.env = env
        o, info = self.env.reset()
        self.root_observation = info[constants.COMMON.OBSERVATION]
        self.gamma = gamma
        self.c = c
        self.c2 = c2
        self.planning_time = planning_time
        self.max_particles = max_particles
        self.reinvigorated_particles_ratio = reinvigorated_particles_ratio
        self.rollout_policy = rollout_policy
        self.initial_visit_count = 0
        self.value_function = value_function
        self.initial_particles = initial_particles
        self.reinvigoration = reinvigoration
        self.default_node_value = default_node_value
        self.prior_weight = prior_weight
        root_particles = copy.deepcopy(self.initial_particles)
        self.tree = BeliefTree(root_particles=root_particles, default_node_value=self.default_node_value,
                               root_observation=self.root_observation, initial_visit_count=self.initial_visit_count)
        self.verbose = verbose
        self.prior_confidence = prior_confidence
        self.acquisition_function_type = acquisition_function_type
        self.use_rollout_policy = use_rollout_policy
        self.prune_action_space = prune_action_space
        self.prune_size = prune_size
        self.num_simulation_steps = 0

    def compute_belief(self) -> Dict[int, float]:
        """
        Computes the belief state based on the particles

        :return: the belief state
        """
        belief_state = {}
        particle_distribution = POMCPUtil.convert_samples_to_distribution(self.tree.root.particles)
        for state, prob in particle_distribution.items():
            belief_state[state] = round(prob, 6)
        return belief_state

    def rollout(self, state: int, history: List[int], depth: int, max_rollout_depth: int, t: int) -> float:
        """
        Perform randomized recursive rollout search starting from the given history
        until the max depth has been achieved

        :param state: the initial state of the rollout
        :param history: the history of the root node
        :param depth: current planning horizon
        :param max_rollout_depth: max rollout depth
        :param t: the time step
        :return: the estimated value of the root node
        """
        if depth > max_rollout_depth:
            if self.value_function is not None:
                o = self.env.get_observation_from_history(history=history)
                return self.value_function(o)
            else:
                return 0
        if not self.use_rollout_policy or self.rollout_policy is None or self.env.is_state_terminal(state):
            rollout_actions = self.env.get_actions_from_particles(particles=[state], t=len(history),
                                                                  observation=history[-1])
            a = POMCPUtil.rand_choice(rollout_actions)
        else:
            a = self.rollout_policy.action(o=self.env.get_observation_from_history(history=history),
                                           deterministic=False)
        self.env.set_state(state=state)
        _, r, _, _, info = self.env.step(a)
        self.num_simulation_steps += 1
        s_prime = info[constants.COMMON.STATE]
        o = info[constants.COMMON.OBSERVATION]
        return float(r) + self.gamma * self.rollout(state=s_prime, history=history + [a, o], depth=depth + 1,
                                                    max_rollout_depth=max_rollout_depth, t=t)

    def simulate(self, state: int, max_rollout_depth: int, c: float, history: List[int], t: int,
                 max_planning_depth: int, depth=0, parent: Union[None, BeliefNode, ActionNode] = None) \
            -> Tuple[float, int]:
        """
        Performs the POMCP simulation starting from a given belief node and a sampled state

        :param state: the sampled state from the belief state of the node
        :param max_rollout_depth: the maximum depth of rollout
        :param max_planning_depth: the maximum depth of planning
        :param c: the weighting factor for the ucb acquisition function
        :param depth: the current depth of the simulation
        :param history: the current history (history of the start node plus the simulation history)
        :param parent: the parent node in the tree
        :param t: the time-step
        :return: the Monte-Carlo value of the node and the current depth
        """

        # Check if we have reached the maximum depth of the tree
        if depth > max_planning_depth:
            if len(history) > 0 and self.value_function is not None:
                o = self.env.get_observation_from_history(history=history)
                return self.value_function(o), depth
            else:
                return 0, depth

        # Check if the new history has already been visited in the past of should be added as a new node to the tree
        observation = -1
        if len(history) > 0:
            observation = history[-1]
        if observation != -1 and self.value_function is not None:
            value = self.value_function(self.env.get_observation_from_history(history))
        else:
            value = self.default_node_value
        current_node = self.tree.find_or_create(history=history, parent=parent, observation=observation,
                                                initial_visit_count=self.prior_confidence, initial_value=value)
        # If a new node was created, then it has no children, in which case we should stop the search and
        # do a Monte-Carlo rollout with a given base policy to estimate the value of the node
        if len(current_node.children) == 0:
            # Prune action space
            if self.prune_action_space and self.rollout_policy is not None:
                obs_vector = self.env.get_observation_from_history(current_node.history)
                dist = self.rollout_policy.model.policy.get_distribution(
                    obs=torch.tensor([obs_vector]).to(self.rollout_policy.model.device)).log_prob(
                    torch.tensor(self.A).to(self.rollout_policy.model.device)).cpu().detach().numpy()
                dist = list(map(lambda i: (math.exp(dist[i]), self.A[i]), list(range(len(dist)))))
                rollout_actions = list(map(lambda x: x[1], sorted(dist, reverse=True,
                                                                  key=lambda x: x[0])[:self.prune_size]))
            else:
                rollout_actions = self.A
                if isinstance(current_node, BeliefNode):
                    rollout_actions = self.env.get_actions_from_particles(particles=current_node.particles + [state],
                                                                          t=t + depth, observation=observation)

            if len(current_node.children) == 0:
                # since the node does not have any children, we first add them to the node
                for action in rollout_actions:
                    self.tree.add(history + [action], parent=current_node, action=action, value=self.default_node_value)

            # Perform the rollout from the current state and return the value
            self.env.set_state(state=state)
            return (self.rollout(state=state, history=history, depth=depth, max_rollout_depth=max_rollout_depth, t=t),
                    depth)

        # If we have not yet reached a new node, we select the next action according to the acquisiton function
        if self.acquisition_function_type == POMCPAcquisitionFunctionType.UCB:
            random.shuffle(current_node.children)
            next_action_node = sorted(
                current_node.children, key=lambda x: POMCPUtil.ucb_acquisition_function(x, c=c), reverse=True)[0]
        elif self.acquisition_function_type == POMCPAcquisitionFunctionType.ALPHA_GO:
            o = self.env.get_observation_from_history(current_node.history)
            if self.rollout_policy is None:
                raise ValueError("Cannot use the AlphaGo acquisiton function without a rollout policy")
            dist = self.rollout_policy.model.policy.get_distribution(
                obs=torch.tensor([o]).to(self.rollout_policy.model.device)).log_prob(
                torch.tensor(self.A).to(self.rollout_policy.model.device)).cpu().detach().numpy()
            dist = list(map(lambda i: math.exp(dist[i]), list(range(len(dist)))))
            next_action_node_idx = sorted(
                list(range(len(current_node.children))), key=lambda x: POMCPUtil.alpha_go_acquisition_function(
                    current_node.children[x], c=c, c2=self.c2, prior=dist[x],
                    prior_weight=self.prior_weight), reverse=True)[0]
            next_action_node = current_node.children[next_action_node_idx]
        else:
            raise ValueError(f"Acquisition function type: {self.acquisition_function_type} not recognized")

        # Simulate the outcome of the selected action
        a = next_action_node.action
        self.env.set_state(state=state)
        _, r, _, _, info = self.env.step(a)
        self.num_simulation_steps += 1
        o = info[constants.COMMON.OBSERVATION]
        s_prime = info[constants.COMMON.STATE]

        # Recursive call, continue the simulation from the new node
        assert isinstance(next_action_node, ActionNode)
        R, rec_depth = self.simulate(
            state=s_prime, max_rollout_depth=max_rollout_depth, depth=depth + 1,
            history=history + [next_action_node.action, o],
            parent=next_action_node, c=c, max_planning_depth=max_planning_depth, t=t)
        R = float(r) + self.gamma * R

        # The simulation has completed, time to backpropagate the values
        # We start by updating the belief particles and the visit count of the current belief node
        if isinstance(current_node, BeliefNode):
            current_node.particles += [state]

        # Next we update the statistics and visit counts of the action node
        if isinstance(next_action_node, ActionNode):
            next_action_node.update_stats(immediate_reward=r)
        current_node.visit_count += 1
        next_action_node.visit_count += 1
        next_action_node.value += (R - next_action_node.value) / next_action_node.visit_count

        return float(R), rec_depth

    def solve(self, max_rollout_depth: int, max_planning_depth: int, t: int) -> None:
        """
        Runs the POMCP algorithm with a given max depth for the lookahead

        :param max_rollout_depth: the max depth for rollout
        :param max_planning_depth: the max depth for planning
        :return: None
        """
        if self.verbose:
            Logger.__call__().get_logger().info(
                f"Starting POMCP, max rollout depth: {max_rollout_depth}, max planning depth: {max_planning_depth}, "
                f"c: {self.c}, planning time: {self.planning_time}, gamma: {self.gamma}, "
                f"max particles: {self.max_particles}, "
                f"reinvigorated_particles_ratio: {self.reinvigorated_particles_ratio}, "
                f"prior weight: {self.prior_weight}, "
                f"acquisiton: {self.acquisition_function_type.value}, c2: {self.c2}, prune: {self.prune_action_space}, "
                f"prune size: {self.prune_size}, use_rollout_policy: {self.use_rollout_policy}, "
                f"prior confidence: {self.prior_confidence}")
        self.num_simulation_steps = 0
        begin = time.time()
        n = 0
        state = self.tree.root.sample_state()
        self.env.set_state(state)
        while time.time() - begin < self.planning_time:
            n += 1
            state = self.tree.root.sample_state()
            _, depth = self.simulate(state=state, max_rollout_depth=max_rollout_depth, history=self.tree.root.history,
                                     c=self.c,
                                     parent=self.tree.root, max_planning_depth=max_planning_depth, depth=0, t=t)
            if self.verbose:
                action_values = np.zeros((len(self.A),))
                best_action_idx = 0
                best_value = -np.inf
                counts = []
                values = []
                for i, action_node in enumerate(self.tree.root.children):
                    action_values[action_node.action] = action_node.value - 10 / (1 + action_node.visit_count)
                    if action_values[action_node.action] > best_value:
                        best_action_idx = i
                        best_value = action_values[action_node.action]
                    counts.append(action_node.visit_count)
                    values.append(round(action_values[action_node.action], 1))
                Logger.__call__().get_logger().info(
                    f"Planning time left {self.planning_time - time.time() + begin}s, "
                    f"best action: {self.tree.root.children[best_action_idx].action}, "
                    f"value: {self.tree.root.children[best_action_idx].value}, "
                    f"count: {self.tree.root.children[best_action_idx].visit_count}, "
                    f"planning depth: {depth}, counts: "
                    f"{sorted(counts, reverse=True)[0:5]}, values: {sorted(values, reverse=True)[0:5]}")
        if self.verbose:
            Logger.__call__().get_logger().info(f"Planning complete, num simulation steps: {self.num_simulation_steps}")

    def get_action(self) -> int:
        """
        Gets the next action to execute based on the state of the tree. Selects the action with the highest value
        from the root node.

        :return: the next action
        """
        root = self.tree.root
        action_vals = [(action.value - 10 / (action.visit_count + 1), action.action) for action in root.children]
        if self.verbose:
            for a in root.children:
                Logger.__call__().get_logger().info(f"action: {a.action}, value: {a.value}, "
                                                    f"visit count: {a.visit_count}")
        return int(max(action_vals)[1])

    def update_tree_with_new_samples(self, action_sequence: List[int], observation: int, t: int) -> List[Any]:
        """
        Updates the tree after an action has been selected and a new observation been received

        :param action_sequence: the action sequence that was executed
        :param observation: the observation that was received
        :param t: the time-step
        :return: the updated particle state
        """
        root = self.tree.root
        if len(action_sequence) == 0:
            raise ValueError("Invalid action sequence")
        action = action_sequence[-1]

        # Since we executed an action we advance the tree and update the root to the the node corresponding to the
        # action that was selected
        child = root.get_child(action)
        if child is not None:
            new_root = child.get_child(observation)
        else:
            raise ValueError(f"Could not find child node for action: {action}")

        # If we did not have a node in the tree corresponding to the observation that was observed, we select a random
        # belief node to be the new root (note that the action child node will always exist by definition of the
        # algorithm)
        if new_root is None:
            # Get the action node
            action_node = root.get_child(action)
            if action_node is None:
                raise ValueError("Chould not find the action node")

            if action_node.children:
                # If the action node has belief state nodes, select a random of them to be the new root
                new_root = POMCPUtil.rand_choice(action_node.children)
            else:
                # or create the new belief node randomly
                particles = copy.deepcopy(self.initial_particles)
                if self.value_function is not None:
                    # initial_value = self.value_function(observation)
                    initial_value = self.default_node_value
                else:
                    initial_value = self.default_node_value
                new_root = self.tree.add(history=action_node.history + [observation], parent=action_node,
                                         observation=observation, particle=particles, value=initial_value,
                                         initial_visit_count=self.prior_confidence)

        # Check how many new particles are left to fill
        if isinstance(new_root, BeliefNode):
            new_root.particles = []
            particle_slots = self.max_particles - len(new_root.particles)
        else:
            raise ValueError("Invalid root node")
        if particle_slots > 0:
            if self.verbose:
                Logger.__call__().get_logger().info(f"Filling {particle_slots} particles")
            particles = []
            # fill particles by Monte-Carlo using reject sampling
            count = 0
            while len(particles) < particle_slots:
                s = root.sample_state()
                self.env.set_state(state=s)
                _, r, _, _, info = self.env.step(action)
                s_prime = info[constants.COMMON.STATE]
                o = info[constants.COMMON.OBSERVATION]
                if o == observation:
                    particles.append(s_prime)
                    count = 0
                else:
                    count += 1
                if count >= 50000:
                    target = root.sample_state().red_agent_target
                    Logger.__call__().get_logger().info(
                        f"Invalid observation: {observation}, target: {target}, "
                        f"given state: 1: \n{root.sample_state()}, \n"
                        f"2: \n {root.sample_state()}\n, 3: {root.sample_state()}\n ")
                    for i in range(particle_slots - len(particles)):
                        s = root.sample_state()
                        self.env.set_state(state=s)
                        _, r, _, _, info = self.env.step(action)
                        s_prime = info[constants.COMMON.STATE]
                        particles.append(s_prime)

            new_root.particles += particles

        # We now prune the old root from the tree
        self.tree.prune(root, exclude=new_root)
        # and update the root
        self.tree.root = new_root

        # Prune children
        if self.verbose:
            Logger.__call__().get_logger().info("Pruning children")
        feasible_actions = (
            self.env.get_actions_from_particles(particles=self.tree.root.particles, t=t, observation=observation,
                                                verbose=True))
        if self.verbose:
            Logger.__call__().get_logger().info(f"feasible actions: {feasible_actions}")
        children = []
        for ch in self.tree.root.children:
            if ch.action in feasible_actions:
                children.append(ch)
        self.tree.root.children = children

        # To avoid particle deprivation (i.e., that the algorithm gets stuck with the wrong belief)
        # we do particle reinvigoration here
        if self.reinvigoration and len(self.initial_particles) > 0:
            if self.verbose:
                Logger.__call__().get_logger().info("Starting reinvigoration")

            # Generate some new particles randomly
            num_reinvigorated_particles = int(len(new_root.particles) * self.reinvigorated_particles_ratio)
            reinvigorated_particles = []
            for i in range(num_reinvigorated_particles):
                s = root.sample_state()
                self.env.set_state(state=s)
                _, r, _, _, info = self.env.step(action)
                s_prime = info[constants.COMMON.STATE]
                reinvigorated_particles.append(s_prime)

            # Randomly exchange some old particles for the new ones
            for particle in reinvigorated_particles:
                new_root.particles[np.random.randint(0, len(new_root.particles))] = particle
        return new_root.particles



from typing import Dict, Any
import numpy.typing as npt
import GPy
from emukit.model_wrappers.gpy_model_wrappers import GPyModelWrapper
from csle_agents.agents.bayesian_optimization_emukit.bo.kernel.kernel_config import KernelConfig


class GPConfig:
    """
    DTO class representing the configuration of a Gaussian Process (GP) based on GPy
    """

    def __init__(self, kernel_config: KernelConfig, obs_likelihood_variance: float = 1e-10) -> None:
        """
        Initializes the DTO

        :param kernel_config: the kernel config
        :param obs_likelihood_variance: the likelihood model of the variance when sampling from the true function
        """
        self.kernel_config = kernel_config
        self.obs_likelihood_variance = obs_likelihood_variance

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"kernel_config: {self.kernel_config}, obs_Likelihood_variance: {self.obs_likelihood_variance}"

    def create_gp(self, X: npt.NDArray[Any], Y: npt.NDArray[Any], input_dim: int) -> GPyModelWrapper:
        """
        Creates the GP model

        :param X: the initial X values
        :param Y: the initial Y values
        :param input_dim: the dimension of the X-values
        :return: the GP model (wrapped in emukit object)
        """
        return GPyModelWrapper(GPy.models.GPRegression(
            X, Y, self.kernel_config.create_kernel(input_dim=input_dim, var_function=None),
            noise_var=self.obs_likelihood_variance))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "GPConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        from csle_agents.agents.bayesian_optimization_emukit.bo.kernel.rbf_kernel_config import RBFKernelConfig
        dto = GPConfig(
            kernel_config=RBFKernelConfig.from_dict(d["kernel_config"]),
            obs_likelihood_variance=d["obs_likelihood_variance"]
        )
        return dto

    def to_dict(self) -> Dict[str, Any]:
        """
        Gets a dict representation of the object

        :return: A dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["obs_likelihood_variance"] = self.obs_likelihood_variance
        d["kernel_config"] = self.kernel_config.to_dict()
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "GPConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return GPConfig.from_dict(json.loads(json_str))



import pytest_mock
from csle_common.dao.training.agent_type import AgentType
from csle_agents.agents.dynasec.dynasec_agent import DynaSecAgent


class TestDynaSecSuite:
    """
    Test suite for the DynaSecAgent
    """

    def test_create_agent(self, mocker: pytest_mock.MockFixture) -> None:
        """
        Tests creation of the FPAgent

        :return: None
        """
        emulation_executions = [mocker.MagicMock()]
        attacker_sequence = mocker.MagicMock()
        defender_sequence = mocker.MagicMock()
        system_id_config = mocker.MagicMock()
        experiment_config = mocker.MagicMock()
        experiment_config.configure_mock(**{"agent_type": AgentType.DYNA_SEC})
        simulation_env_config = mocker.MagicMock()
        DynaSecAgent(emulation_executions=emulation_executions, simulation_env_config=simulation_env_config,
                     experiment_config=experiment_config, attacker_sequence=attacker_sequence,
                     defender_sequence=defender_sequence, system_identification_config=system_id_config)
